/*
 * ssheet.c
 *
 * definition of the languages style sheets
 * Copyright (c) 1988, 89, 90, 91, 92, 93 Miguel Santana
 * Copyright (c) 1995, 96, 97 Akim Demaille, Miguel Santana
 * $Id: ssheet.c,v 1.74 1997/11/07 11:35:43 demaille Exp $
 */

/*
 * This file is part of a2ps.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "a2ps.h"
#include "routines.h"
#include "ssheet.h"
#include "pathwalk.h"
#include "jobs.h"
#include "msg.h"
#include "filtdir.h"
#include "xstrrpl.h"
#include "xalloca.h"

#ifndef HAVE_STRCOLL
#  define strcoll strcmp
#endif

/* Suffix of the style sheet files */
#define SSH_SUFFIX "ssh"

/* The rules to convert the sheet's documentation to another format */
/* 1. Plain ASCII */
#define DOC_TO_PLAIN \
  "url(", "",		\
  ")url(", " (",	\
  ")url", ")",		\
  "@example\n",		"",	\
  "@end example\n", 	"",	\
  "@end example", 	"",	/* Just in case */	\
  "@itemize\n",		"",	\
  "@end itemize", 	"",	\
  "@item\n",		" - ",	\
  NULL

#define DOC_TO_HTML \
  "url(", "<a href=\"",	\
  ")url(", "\">",	\
  ")url", "</a>",	\
  "@example",		"<pre>",	\
  "@end example", 	"</pre>",	\
  "@itemize",		"<ul>",	\
  "@end itemize", 	"</ul>",	\
  "@item\n",		"<li>",	\
  NULL

#define DOC_TO_TEXINFO \
  "url(", "@href{",	\
  ")url(", ",",		\
  ")url", "}",		\
  "@itemize",		"@itemize @minus",	\
  NULL

extern a2ps_job * job;
extern struct hash_table * style_sheets;

/* Comes from regex */
extern char * re_syntax_table;

/* This function is defined in sshparser.y, but I don't know
 * how to make it appear in a .h file */
int parse_style_sheet PARAMS ((const char * filename));

/* Computes the sum of the versions and the highest requirement
 * of the sheet with key in KEYS */
static inline void style_sheets_versions PARAMS ((struct darray * keys,
						  int sum [4],
						  int requirement [4]));

/* return TRUE iff OK */
static int style_sheet_check PARAMS ((struct style_sheet * sheet));

/* We drop in the alphabets[0] a special value, which cannot
 * be defined by the definition of the alphabets.  It allows
 * to see if the alphabets were defined or not, so that
 * the default alphabets can be set when finalizing the sheet.
 * Note that we cannot set the default alphabets a priori,
 * because we need to be able to inherit from ancestors
 */
#define ALPHABET_UNDEFINED_TAG	127
#define ALPHABET_IS_UNDEFINED(_alpha_)	\
	((_alpha_)[0] == ALPHABET_UNDEFINED_TAG)
#define UNDEFINE_ALPHABET(_alpha_)	\
	((_alpha_)[0] = ALPHABET_UNDEFINED_TAG)

#define DEFAULT_ALPHA1 ((uchar *) \
	"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_")
#define DEFAULT_ALPHA2 ((uchar *) \
	"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789")

/* Same story, with the sensitivity */
#define SENSITIVENESS_UNDEFINED_TAG	127
#define SENSITIVENESS_IS_UNDEFINED(_sens_)	\
	((_sens_) == SENSITIVENESS_UNDEFINED_TAG)

#define DEFAULT_SENSITIVENESS CASE_INSENSITIVE

struct pattern *
new_pattern (char * pattern, size_t len)
{
  struct pattern * res = ALLOC (struct pattern, 1);
  res->pattern = pattern;
  res->len = len;
  return res;
}

/*
 * The faced_string
 */
struct faced_string *
faced_string_new (uchar * string, int reg_ref, face_t face)
{
  struct faced_string * res = ALLOC (struct faced_string, 1);
  res->string = string;
  res->reg_ref = reg_ref;
  res->face = face;
  return res;
}

/*
 * Report a faced_string
 */
static void
faced_string_self_print (FILE * stream, struct faced_string * faced_string)
{
  uchar * facename = xface_to_string (faced_string->face);

  if (faced_string->string)
    fprintf (stream, "\"%s\": %s", faced_string->string, facename);
  else 
    fprintf (stream, "\\%d: %s", faced_string->reg_ref, facename);

  FREE (facename);
}

inline static void
faced_string_free (struct faced_string * faced_string)
{
  XFREE (faced_string->string);
  FREE (faced_string);
}

/*
 * Replace all the No_face's by FACE
 */
inline static void
faced_string_set_no_face (struct faced_string * str, face_t face)
{
  if (str->face == No_face)
    str->face = face;
}

/************************************************************************/
/*	The version handling						*/
/************************************************************************/
void
style_sheet_set_version (struct style_sheet * sheet,
			 const char * version_string)
{
  string_to_version (version_string, sheet->version);
}

/*
 * Set the version of a2ps required for the processing of SHEET.
 * Return FALSE if the requirement is higher than VERSION.
 */
int
style_sheet_set_requirement (struct style_sheet * sheet,
			     const char * version_string)
{
  int a2ps_version [4];

  string_to_version (version_string, sheet->requirement);
  string_to_version (VERSION, a2ps_version);

  if (version_cmp (sheet->requirement, a2ps_version) > 0)
    return FALSE;
  
  return TRUE;
}

/************************************************************************/
/*	Destination darray						*/
/************************************************************************/
/*
 * New destinations array
 */
struct darray *
destinations_new (void)
{
  struct darray * res;
  res = da_new ("Destinations", 2, da_geometrical, 2,
		(da_print_func_t) faced_string_self_print, NULL);
  return res;
}

/*
 * New destinations array with a single argument
 */
struct darray *
destinations_new_single (uchar * string, int reg_ref, face_t face)
{
  struct darray * res;
  res = da_new ("Destinations", 2, da_geometrical, 2,
		(da_print_func_t) faced_string_self_print, NULL);
  da_append (res, faced_string_new (string, reg_ref, face));
  return res;
}

/*
 * Replace all the No_face's by FACE
 */
inline static void
destinations_set_no_face (struct darray * dest, face_t face)
{
  int i;

  for (i = 0 ; i < dest->len ; i++)
    if (((struct faced_string *) dest->content [i])->face == No_face)
      ((struct faced_string *) dest->content [i])->face = face;
}

/*
 * Add a symbol in the destination array,
 * Don't forget to scan the string, because it may have actually
 * regex references in it, which must be resolved
 */
void
destinations_add (struct darray * dest, struct faced_string * str)
{
  da_append (dest, str);
}

/*
 * Report a destination array
 */
void
destinations_self_print (FILE * stream, struct darray * destinations)
{
  da_dump_content (destinations, stream);
}

/************************************************************************/
/*	Symbols								*/
/************************************************************************/
/*
 * Dealing with the symbols and operators
 */
inline static struct symbol *
new_symbol_internal_word (uchar * word,
			  struct darray * destination)
{
  struct symbol * res = ALLOC (struct symbol, 1);

  res->word = word;
  res->regex = NULL;
  res->destination = destination;
  return res;
}

/*
 * Dealing with the symbols and operators
 */
inline static struct symbol *
new_symbol_internal_regexp (struct pattern * pattern,
			    char * pattern_to_compile, int pattern_len,
			    struct darray * destination)
{
  struct symbol * res = ALLOC (struct symbol, 1);
  const char * error_msg;
  /* This is a regular expression.  We want to keep the original
   * pattern to ease the debugging of a style sheet */
  res->word = (uchar *) pattern->pattern;
  
  /* Build the regex structure, and compile the pattern */
  res->regex = ALLOC (struct re_pattern_buffer, 1);
  res->regex->translate = NULL;
  res->regex->fastmap = NULL;
  res->regex->buffer = NULL;
  res->regex->allocated = 0;
  
  error_msg = re_compile_pattern (pattern_to_compile, pattern_len,
				  res->regex);
  /*
    fprintf (stderr, "Found %d subexp in `%s'\n",
    res->regex->re_nsub, pattern->pattern);
  */
  if (error_msg)
    error (1, 0, _("cannot compile regular expression: %s"),
	   error_msg);
  FREE (pattern);
  res->destination = destination;
  return res;
}

/*
 * Dealing with the symbols and operators
 */
struct symbol *
new_symbol (uchar * word, struct pattern * pattern,
	    struct darray * destination)
{
  if (pattern)
    return new_symbol_internal_regexp (pattern,
				       pattern->pattern, pattern->len,
				       destination);
  else 
    return new_symbol_internal_word (word,
				     destination);
}

/*
 * Dealing with the symbols and operators
 * This is pretty much the same as the previous case, 
 * but the difference is that the regexp much be built appending
 * a \\b at the end, to make sure that, for instance,
 * as a keyword matcher /c*f/ does not match "ccfe": the keyword
 * is not separated.
 */
struct symbol *
new_keyword_symbol (uchar * word, struct pattern * pattern,
		    struct darray * destination)
{
  int pattern_len = 0;
  char * pattern_to_compile = NULL;

  if (pattern)
    {
      pattern_len = pattern->len + strlen ("\\b\\(\\)\\b");
      pattern_to_compile = ALLOCA (char, pattern_len + 1);
      sprintf (pattern_to_compile, "\\b\\(%s\\)\\b", pattern->pattern);
      return new_symbol_internal_regexp (pattern,
					 pattern_to_compile, pattern_len,
					 destination);
    }
  else
    return new_symbol_internal_word (word,
				     destination);
}


void
free_symbol (struct symbol * symbol)
{
  XFREE (symbol->word);
  if (symbol->regex) {
    regfree (symbol->regex);
    FREE (symbol->regex);
  }
  da_free (symbol->destination, (da_map_func_t) faced_string_free);
  FREE (symbol);
}

int
symbol_cmp (struct symbol * k1, struct symbol * k2)
{
  
  /* A regex is always smaller than a word */
  if (k1->regex && !k2->regex)
    return -1;
  if (!k1->regex && k2->regex)
    return 1;

  /* Then compare upon the original value (even for regex:
   * qsort breaks if the order is not total) */
  return ustrcmp (k1->word, k2->word);
}

void
symbol_self_print (FILE * stream, struct symbol * sym)
{
  if (sym->regex)
    fprintf (stream, "/%s/ -> ", sym->word);
  else
    fprintf (stream, "\"%s\" -> ", sym->word);
  destinations_self_print (stream, sym->destination);
}

/*
 * Replace all the No_face's that may be in the array of
 * symbols, to FACE
 */
inline void
symbol_set_no_face (struct symbol * symbol, face_t face)
{
  destinations_set_no_face (symbol->destination, face);
}

void
symbols_set_no_faces (struct darray * arr, face_t face)
{
  int i;
  
  for (i = 0 ; i < arr->len ; i++)
    symbol_set_no_face (((struct symbol *) arr->content[i]), face);
}


/*
 * `words' are a specialization of darray, to capture the fact
 * that it is faster to look for a prefix of a string taking
 * advantage to know that it can only be in the interval
 * of candidates that start with the same first letter
 *
 */
static struct words *
words_create (/* Regular darray parameters */
	      const char * name, size_t size, 
	      size_t increment)
{
  int c;
  char * name1, *name2;
  struct words * res = ALLOC (struct words, 1);
  
  /* Initialize the underlying darray of strings */
  name1 = MALLOC (strlen (name) + strlen (": strings") + 1);
  sprintf (name1, "%s: strings", name);
  res->strings = da_new (name1, size, 
			 da_linear, increment, 
			 (da_print_func_t) symbol_self_print,
			 (da_cmp_func_t) symbol_cmp);
  
  /* Initialize the underlying darray of regexps */
  name2 = MALLOC (strlen (name) + strlen (": regexps") + 1);
  sprintf (name2, "%s: regexps", name);
  res->regexps = da_new (name2, size, 
			 da_linear, increment, 
			 (da_print_func_t) symbol_self_print,
			 (da_cmp_func_t) symbol_cmp);
  
  /* Set min and max intervals to void values */
  for (c = 0 ; c < 256 ; c ++) {
    res->min[c] = NULL;
    res->max[c] = NULL;
  }

  return res;
}

static void
words_dump_content (struct words * words, FILE * stream)
{
  da_dump_content (words->regexps, stream);
  da_dump_content (words->strings, stream);
}

/*
 * Add the content of NEW in WORDS.
 * If in WORDS there is yet a symbol with the same string matcher
 *    than in NEW, then the one of NEW is the winner
 * No special care is made for the regex matcher
 */
void
words_merge_symbols_unique (struct words * words, struct darray * new)
{
  int i;

  /* 1. Select those that are regexps */
  i = 0;
  while (i < new->len) {
    if (((struct symbol *) new->content [i])->regex) {
      /* This is a regexp */
      da_append (words->regexps, ((struct symbol *) new->content [i]));
      /* Don't free it, since its content is shared */
      da_remove_at (new, i, NULL);
    } else
      i++;
  }

  /* 2. Include the strings */
  da_include_unique (words->strings, new,
		     (da_map_func_t) free_symbol);
}

/*
 * Finish setting up the keywords recognition mechanism
 */
static void
words_finalize (struct words * list)
{
  int i;
  struct symbol ** content;

  /* Sort in alphabetical order */
  da_qsort (list->strings);
  
  /* Build two arrays which contains for each letter, its first
   * and last occurence index in (*DEST) array */
  content = (struct symbol **) list->strings->content;
  for (i = 0 ; i < list->strings->len ; i++) {
    if (list->min [content[i]->word[0]] == NULL)
      list->min [content[i]->word[0]] = content + i;
    list->max [content[i]->word[0]] = content + i;
  }
}

/************************************************************************/
/*		Handling the ancestors of a style sheet 		*/
/************************************************************************/
static void
ancestors_finalize (struct style_sheet * sheet)
{
  int i;
  struct style_sheet * ancestor = NULL;
  
  if (sheet->ancestors->len == 0)
    return;	/* No inheritance */

  /* First take the ancestors in the given order */
  for (i = 0 ; i < sheet->ancestors->len ; i++) {
    ancestor = get_style_sheet ((char *) sheet->ancestors->content [i]);
    /* Inherit from their keywords */
    da_include_unique_conservative (sheet->keywords->strings, 
				    ancestor->keywords->strings, 
				    (da_map_func_t) free_symbol);
    da_concat (sheet->keywords->regexps,
	       ancestor->keywords->regexps);
    
    /* Inherit from their operators */
    da_include_unique_conservative (sheet->operators->strings, 
				    ancestor->operators->strings, 
				    (da_map_func_t) free_symbol);
    da_concat (sheet->operators->regexps,
	       ancestor->operators->regexps);
  }
  
  /* Inherit of the _last_ alphabets if it has not been defined in
   * this sheet.  */
  if (ALPHABET_IS_UNDEFINED (sheet->alpha1))
    memcpy ((POINTER) sheet->alpha1, 
	    (POINTER) ancestor->alpha1, sizeof (uchar) * 256);
  if (ALPHABET_IS_UNDEFINED (sheet->alpha2))
    memcpy ((POINTER) sheet->alpha2,
	    (POINTER) ancestor->alpha2, sizeof (uchar) * 256);

  /* Inherit of the _last_ case sensitivity */
  if (SENSITIVENESS_IS_UNDEFINED (sheet->sensitiveness))
    sheet->sensitiveness = ancestor->sensitiveness;

  /* Then, in reverse order */
  for (i = sheet->ancestors->len - 1; i >= 0 ; i--) {
    ancestor = get_style_sheet ((char *) sheet->ancestors->content [i]);
    
    /* Inherit from their sequences */
    /* The must be read in reverse order */
    da_prefix (sheet->sequences, ancestor->sequences);
  }
}

/************************************************************************/
/*		Creating, loading a style sheet				*/
/************************************************************************/
/*
 * Create a sequence
 */
struct sequence *
new_sequence (struct symbol * Open,
	      face_t in_face,
	      struct darray * Close,
	      struct darray * exceptions)
{
  struct sequence * res = ALLOC (struct sequence, 1);

  /* Make sure to set the face of those with No_face */
  symbol_set_no_face (Open, in_face);
  res->open = Open;

  res->face = in_face;

  symbols_set_no_faces (exceptions, in_face);
  res->exceptions = exceptions;

  symbols_set_no_faces (Close, in_face);
  res->close = Close;
  return res;
}

/*
 * Release a sequence.
 */
void
free_sequence (struct sequence * sequence)
{
  free_symbol (sequence->open);
  da_free (sequence->close, (da_map_func_t) free_symbol);
  da_free (sequence->exceptions, (da_map_func_t) free_symbol);
  FREE (sequence);
}

/*
 * Build the usual C's string and char with their escapes
 */
#define C_add_exception(_excep_)					\
  da_append (res, 							\
	     new_symbol (xustrdup (_excep_), NULL,			\
			 destinations_new_single (xustrdup (_excep_), 0,\
						 String)));
static struct darray *
new_C_exceptions (void)
{
  struct darray * res = da_new ("C usual exceptions", 3, 
				da_linear, 2,
				(da_print_func_t) symbol_self_print,
				(da_cmp_func_t) symbol_cmp);
  C_add_exception ("\\\'");
  C_add_exception ("\\\\");
  C_add_exception ("\\\"");
  return res;
}

struct sequence *
new_C_string_sequence (const char * delimitor)
{
  struct sequence * res = ALLOC (struct sequence, 1);
  res->open = new_symbol ((uchar *) delimitor, NULL,
			  destinations_new_single (NULL, 0, Plain));
  res->face = String;
  res->exceptions = new_C_exceptions ();
  res->close = da_new ("Closing alternatives", 5,
		       da_linear, 5,
		       (da_print_func_t) symbol_self_print, NULL);
  da_append (res->close, 
	     new_symbol ((uchar *) delimitor, NULL,
			 destinations_new_single (NULL, 0, Plain)));
  return res;
}

void
sequence_self_print (FILE * stream, struct sequence * tmp)
{
  uchar * facename = xface_to_string (tmp->face);

  symbol_self_print (stream, tmp->open);
  fprintf (stream, " -> %s:\n", facename);
  if (tmp->exceptions->len)
    da_dump_content (tmp->exceptions, stream);
  da_dump_content (tmp->close, stream);
  FREE (facename);
}

/************************************************************************/
/*		Creating, loading a style sheet				*/
/************************************************************************/
/*
 * Dealing with the style sheets
 */
struct style_sheet *
new_style_sheet (const uchar * name)
{
  struct style_sheet * res = ALLOC (struct style_sheet, 1);

  res->name = name;
  res->author = (const uchar *) "";
  version_set_to_null (res->version);
  res->comment = UNULL;
  version_set_to_null (res->requirement);
  res->ancestors = da_new ("Ancestors", 2, 
			   da_linear, 2,
			   (da_print_func_t) da_str_print, NULL);

  res->sensitiveness = SENSITIVENESS_UNDEFINED_TAG;

  UNDEFINE_ALPHABET (res->alpha1);
  UNDEFINE_ALPHABET (res->alpha2);

  res->keywords = 
    words_create ("Keywords", 100, 100);

  res->operators = 
    words_create ("Operators", 100, 100);

  res->sequences = da_new ("Sequences", 100, 
			   da_linear, 100,
			   (da_print_func_t) sequence_self_print, NULL);

  return res;
}

/*
 * Once a style sheet has been read, make it usable (sorts etc.)
 * NOTE: It must _not_ be called from the parser, since
 *       it may call another time the parser, in case one of
 *       the ancestors are not known.
 */
void
style_sheet_finalize (struct style_sheet * sheet)
{
  message (msg_sheet3, "Finalizing %s.ssh (%s)\n", sheet->key, sheet->name);

  ancestors_finalize (sheet);

  /* Ensure a default alphabet */
  if (ALPHABET_IS_UNDEFINED (sheet->alpha1))
    string_to_array (sheet->alpha1, DEFAULT_ALPHA1);
  if (ALPHABET_IS_UNDEFINED (sheet->alpha2))
    string_to_array (sheet->alpha2, DEFAULT_ALPHA2);

  /* Ensure a default sensitivity */
   if (SENSITIVENESS_IS_UNDEFINED (sheet->sensitiveness))
     sheet->sensitiveness = DEFAULT_SENSITIVENESS;

  words_finalize (sheet->keywords);
  words_finalize (sheet->operators);
  if (message_test (msg_sheet3)) {
    message (msg_sheet3, "Checking coherence of `%s'\n", sheet->key);
    if (style_sheet_check (sheet))
      message (msg_sheet3, "`%s' is sane.\n", sheet->key);
    else
      message (msg_sheet3, "`%s' is corrupted.\n", sheet->key);
  }
}

/*
 * Loading a style sheet.  In this precise case, we specify
 * the style sheet file to use.  It is used to check that an ssh
 * file is OK, from a2ps-mode in emacs.
 */
static struct style_sheet *
load_style_sheet_by_file (const char * file)
{
  struct style_sheet * item;
  struct style_sheet token;
  char * key, * last_dir_sep, * last_dot;

  parse_style_sheet (file);
 
  /* Find the key */
  last_dir_sep = strrchr (file, DIR_SEP);
  last_dot = strrchr (last_dir_sep, '.');
  key = MALLOC (last_dot - last_dir_sep);
  strncpy (key, last_dir_sep + 1, last_dot - last_dir_sep - 1);
  key [last_dot - last_dir_sep - 1] = '\0';

  /* Finalize it. For this, we must find back that style sheet
   * which has been put in the hash table by the parser itself */
  /* This cannot fail.  parse_style_sheet would have failed before */
  token.key = key;
  item = (struct style_sheet *) hash_find_item (style_sheets, &token);
  
  style_sheet_finalize (item);
  if (message_test (msg_sheet3))
    style_sheet_self_print (stderr, item);

  return item;
}


/*
 * Loading a style sheet
 */
static struct style_sheet *
load_style_sheet_by_key (const char * key)
{
  char * file;
  struct style_sheet * res;
  
  file = pw_find_file (job->lib_path, key, ".ssh");
  if (!file) {
    error (0, 0, _("cannot find file `%s%s'"), key, ".ssh");
    return 0;
  }
  
  res = load_style_sheet_by_file (file);
  FREE (file);
  return res;
}

/************************************************************************/
/*		Reporting information about the style sheets		*/
/************************************************************************/
/*
 * Compare the names of two style sheets
 * We use the collate comparison, so that `a' is not after `Z'
 */
static int
sheet_name_cmp (const char * key1, const char * key2)
{
  return strcoll ((char *) get_style_sheet (key1) -> name,
		  (char *) get_style_sheet (key2) -> name);
}

/*
 * Report everything (used in -v5 at end of parsing)
 */
void
style_sheet_self_print (FILE * stream, struct style_sheet * sheet)
{
  int c;

  fprintf (stream, "Style sheet for %s (%s.ssh), version ", 
	   sheet->name, sheet->key);
  version_self_print (sheet->version, stream);
  fprintf (stream, ":\nWritten by %s\n", sheet->author);
  if (sheet->ancestors->len > 0) {
    int i;
    fprintf (stream, "It inherits from: ");
    for (i = 0 ; i < sheet->ancestors->len ; i++)
      fprintf (stream, "%s ", (char *) sheet->ancestors->content [i]);
    fputs ("\n", stream);
  }

  fprintf (stream, "Case %ssensitive\n", 
	   sheet->sensitiveness == CASE_INSENSITIVE ? "in" : "");
  fprintf (stream, "First alphabet is \"");
  for (c = 0 ; c < 256 ; c++)
    if (sheet->alpha1 [c])
      fputc (c, stream);
  fprintf (stream, "\"\n");
  fprintf (stream, "Second alphabet is \"");
  for (c = 0 ; c < 256 ; c++)
    if (sheet->alpha2 [c])
      fputc (c, stream);
  fprintf (stream, "\"\n");
  words_dump_content (sheet->keywords, stream);
  words_dump_content (sheet->operators, stream);
  da_dump_content (sheet->sequences, stream);
}

/*
 * Print a short signature (i.e., name, author etc.)
 */
static void
style_sheet_print_signature (FILE * stream, struct style_sheet * sheet)
{
  int i, title_bar_len;

  if (!version_is_null (sheet->version)) {
    title_bar_len = (strlen (" (.ssh version )")
		     + ustrlen (sheet->name)
		     + strlen (sheet->key)
		     + version_length (sheet->version));
    fprintf (stream, "%s (%s.ssh version ", 
	     sheet->name, sheet->key);
    version_self_print (sheet->version, stream);
    fputs (")\n", stream);
  } else {
    title_bar_len = (strlen (" (.ssh)")
		     + ustrlen (sheet->name)
		     + strlen (sheet->key));
    fprintf (stream, "%s (%s.ssh)\n", 
	     sheet->name, sheet->key);
  }
  for (i = 0 ; i < title_bar_len ; i++)
    putc ('-', stream);
  putc ('\n', stream);

  if (!IS_EMPTY(sheet->author)) 
    {
      char * author = xvstrrpl ((char *) sheet->author,
				"Akim Demaille <demaille@inf.enst.fr>", "",
				NULL);
      if (!IS_EMPTY(author))
	fprintf (stream, "Written by %s.\n", author);
      FREE (author);
    }
  
  if (!IS_EMPTY(sheet->comment)) 
    {
      char * documentation =  xvstrrpl ((char *) sheet->comment,
					DOC_TO_PLAIN);
      fprintf (stream, "%s\n", documentation);
      FREE (documentation);
    }
  fputs ("\n", stream);
}

/*
 * List the style sheet names (for option -E)
 */
void
list_style_sheets_short (a2ps_job * the_job, FILE * stream)
{
  struct darray * entries;
  int i;

  entries = pw_glob_on_suffix (job->lib_path, SSH_SUFFIX);

  fputs (_("Known style sheets (languages)"), stream);
    
  for (i = 0 ; i < entries->len ; i++) {
    if (!(i % 7))
      fputs ("\n  ", stream);
    fprintf (stream, "%-10s", (char *) entries->content[i]);
  }
  putc ('\n', stream);
  da_free (entries, NULL);
}

/*
 * List the style sheets
 */
void
list_style_sheets_long (a2ps_job * the_job, FILE * stream)
{
  struct style_sheet * sheet;
  struct darray * entries;
  int i;

  entries = pw_glob_on_suffix (job->lib_path, SSH_SUFFIX);

  /* Sort them by name (not key) */
  entries->cmp = (da_cmp_func_t) sheet_name_cmp;
  da_qsort (entries);

  fputs (_("Known style sheets (languages)"), stream);
  putc ('\n', stream);
  
  for (i = 0 ; i < entries->len ; i++) {
    sheet = get_style_sheet ((char *) entries->content[i]);
    style_sheet_print_signature (stream, sheet);
  }
  putc ('\n', stream);
  da_free (entries, NULL);
}

/************************************************************************/
/*		Report in HTML format					*/
/************************************************************************/
/*
 * Print a short signature (i.e., name, author etc.)
 * In html format
 */
static void
style_sheet_html_print_signature (FILE * stream, struct style_sheet * sheet)
{
  char * author;

  if (version_is_null (sheet->version))
    fprintf (stream, "<li><strong><a href=\"%s.ssh\">%s</a></strong>.</p>\n",
	     sheet->key, sheet->name);
  else {
    fprintf (stream, 
	     "<li><p><strong><a href=\"%s.ssh\">%s</a></strong> version ",
	     sheet->key, sheet->name);
    version_self_print (sheet->version, stream);
    fputs (".</p>\n", stream);      
  }

  if (!IS_EMPTY(sheet->author)
      || !version_is_null(sheet->requirement)
      || (sheet->ancestors->len > 0))
    {
      fputs ("<p>", stream);
      if (!IS_EMPTY(sheet->author)) {
	author = xvstrrpl ((char *) sheet->author,
			   "Akim Demaille <demaille@inf.enst.fr>", "",
			   "<", "(<a href=\"mailto:",
			   ">", "\">mail</a>)",
			   NULL);
	if (!IS_EMPTY(author))
	  fprintf (stream, "Written by %s.  \n", author);
	FREE (author);
      }
      if (!version_is_null (sheet->requirement)) {
	fputs ("It requires a2ps version ", stream);
	version_self_print (sheet->requirement, stream);
	fputs (".\n", stream);
      }
      if (sheet->ancestors->len > 0) {
	int i;
	fputs ("It inherits from: ", stream);
	for (i = 0 ; i < sheet->ancestors->len ; i++)
	  fprintf (stream, "%s.ssh ", (char *) sheet->ancestors->content [i]);
	fputs (".  \n", stream);
      }
      fputs ("</p>", stream);
    }
  
  if (!IS_EMPTY(sheet->comment)) {
    char * documentation = xvstrrpl ((char *) sheet->comment,
				     DOC_TO_HTML);
    fprintf (stream, "<p>%s</p>\n", documentation);
    FREE (documentation);
  }
  fputs ("</li>\n\n", stream);
}

/*
 * List the style sheets in html format
 * I have a conscience problem wrt this.  This is made to ease
 * my task, but actually it has not much to do with a2ps itself :(
 */
void
list_html_style_sheets (a2ps_job * the_job, FILE * stream)
{
  struct style_sheet * sheet;
  struct darray * entries;
  int i, version_index[4], requirement [4];

  entries = pw_glob_on_suffix (job->lib_path, SSH_SUFFIX);

  /* Sort them by name (not key) */
  entries->cmp = (da_cmp_func_t) sheet_name_cmp;
  da_qsort (entries);

  style_sheets_versions (entries, version_index, requirement);

  fputs ("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\n", stream);
  fputs ("<html>\n", stream);
  fputs ("<head>\n", stream);
  fputs ("   <title>a2ps' languages support</title>\n", stream);
  fputs ("</head>\n", stream);
  fputs ("<body background=\"http://www-inf.enst.fr/~demaille/images/chalk.jpg\" bgcolor=\"#FFFFFF\">\n", stream);
  fputs ("\n", stream);
  fputs ("<h1 align=\"CENTER\">\n", stream);
  fprintf (stream, "a2ps' languages support<br>(index = ");
  version_self_print (version_index, stream);
  fputs (")\n", stream);
  fputs ("</h1>\n", stream);
  
  fputs ("<img src=\"http://www-inf.enst.fr/~demaille/images/barre.bleu.bleu.gif\" ALT=\"\" height=4 width=100% align=center>\n", stream);
  fputs ("<h2>Current <a href=\"sheets.mp\">sheets.map</a></h2>\n", stream);
  fputs ("<p>It is the list of rules to decide what style sheet to use.</p>\n", stream);

  fputs ("<img src=\"http://www-inf.enst.fr/~demaille/images/barre.bleu.bleu.gif\" ALT=\"\" height=4 width=100% align=center>\n", stream);

  fputs ("<h2>Current style sheets</h2>\n", stream);
  fprintf (stream, "<p>There is a <a href=\"sheets.tar.gz\">package</a> which contains \
them all.  Its index (sum of the style sheets version number) is \
<strong>");
  version_self_print (version_index, stream);
  fputs ("</strong>, and the highest requirement is a2ps version ", stream);
  version_self_print (requirement, stream);
  fputs (".</p>\n", stream);
  fputs ("<ul>\n", stream);
  for (i = 0 ; i < entries->len ; i++) {
    sheet = get_style_sheet ((char *) entries->content[i]);
    style_sheet_html_print_signature (stream, sheet);
  }
  fputs ("</ul>\n", stream);

  fputs ("<!--#include virtual=\"http://www-inf.enst.fr/~demaille/images/bottom.html\"-->\n", stream);

  fputs ("</body>\n", stream);
  fputs ("</html>\n", stream);

  da_free (entries, NULL);
}

/************************************************************************/
/*		Report in Texinfo format				*/
/************************************************************************/
/*
 * Print a short signature (i.e., name, author etc.)
 * In texinfo format
 */
static void
style_sheet_texinfo_print_signature (FILE * stream,
				     struct style_sheet * sheet)
{
  char * author;

  fprintf (stream, 
	   "@deftp {Style Sheet} {%s} (@file{%s.ssh})\n",
	   sheet->name, sheet->key);

  if (!IS_EMPTY(sheet->author)
      || !version_is_null (sheet->requirement)
      || (sheet->ancestors->len > 0)) {
    if (!IS_EMPTY(sheet->author)) {
      /* 1. We don't want to see tons of demailles
       * 2. We must quote the @ of the emails */
      author = xvstrrpl ((char *) sheet->author,
			 "Akim Demaille <demaille@inf.enst.fr>", "",
			 "<", "@email{",
			 "@", "@@",
			 ">", "}",
			 NULL);
      if (!IS_EMPTY(author))
	fprintf (stream, "Written by %s.  \n", author);
      FREE (author);
    }
    if (!version_is_null (sheet->requirement)) {
      fputs ("It requires a2ps version ", stream);
      version_self_print (sheet->requirement, stream);
      fputs (".  \n", stream);
    }
    if (sheet->ancestors->len > 0) {
      int i;
      fputs ("It inherits from: ", stream);
      for (i = 0 ; i < sheet->ancestors->len ; i++)
	fprintf (stream, "%s@file{%s.ssh}",
		 i ? ", " : "",
		 (char *) sheet->ancestors->content [i]);
      fputs (".  \n", stream);
    }
  }

  if (!IS_EMPTY(sheet->comment)) {
    char * documentation = xvstrrpl ((char *) sheet->comment,
				     DOC_TO_TEXINFO);
    fprintf (stream, "%s\n", documentation);
    FREE (documentation);
  }

  fputs ("@end deftp\n\n", stream);
}

/*
 * List the style sheets in Texinfo format
 */
void
list_texinfo_style_sheets (a2ps_job * the_job, FILE * stream)
{
  struct style_sheet * sheet;
  struct darray * entries;
  int i, version_index[4], requirement[4];

  entries = pw_glob_on_suffix (job->lib_path, SSH_SUFFIX);

  /* Sort them by name (not key) */
  entries->cmp = (da_cmp_func_t) sheet_name_cmp;
  da_qsort (entries);
  
  style_sheets_versions (entries, version_index, requirement);
  
  fprintf (stream, "\
The current index (sum of all the style sheets version number) is ");
  version_self_print (version_index, stream);
  fputs (" and the highest requirement is a2ps version ", stream);
  version_self_print (requirement, stream);
  fputs (".\n\
The known styles are:\n", stream);
  
  for (i = 0 ; i < entries->len ; i++) {
    sheet = get_style_sheet ((char *) entries->content[i]);
    style_sheet_texinfo_print_signature (stream, sheet);
  }

  da_free (entries, NULL);
}

/************************************************************************/
/*		Perform some tests on the style sheets			*/
/************************************************************************/
/*
 * Check that a darray of symbols is in lower case
 * (TRUE for success)
 */
static int
check_symbols_lower_case (const char * name, struct darray * symbols)
{
  int i;
  int res = TRUE;

  for (i = 0 ; i < symbols->len ; i++)
    if (!is_strlower (((struct symbol *) symbols->content[i])->word)) {
      error (0, 0, 
	     "%s.ssh:`%s' uses upper case characters",
	     name, ((struct symbol *) symbols->content[i])->word);
      res = FALSE;
    }
  return res;
}

/*
 * Check that WORDS is all in lower case
 */
static int
check_words_lower_case (const char * name, struct words * words)
{
  return (check_symbols_lower_case (name, words->strings)
	  && check_symbols_lower_case (name, words->regexps));
}

/*
 * Check that SEQUENCE is all in lower case
 */
static int
check_sequence_lower_case (const char * name, struct sequence * sequence)
{
  int res = TRUE;

  if (!is_strlower (sequence->open->word)) {
    error (0, 0, 
	   "%s.ssh:`%s' uses upper case characters",
	   name, sequence->open->word);
    res = FALSE;
  }
  
  res &= check_symbols_lower_case (name, sequence->exceptions);
  res &= check_symbols_lower_case (name, sequence->close);

  return res;
}

static int
check_sequences_lower_case (const char * name, struct darray * sequences)
{
  int i;
  int res = TRUE;

  for (i = 0 ; i < sequences->len ; i++)
    res &= check_sequence_lower_case (name,
				      (struct sequence *) sequences->content [i]);

  return res;
}

/*
 * Check that keywords are in lower case when lang is
 * case insensitive (returns 0 for failure)
 */
static int
check_sensitivity (struct style_sheet * sheet)
{
  int res = TRUE;
  
  if (sheet->sensitiveness == CASE_SENSITIVE)
    return TRUE;

  /* The language is case insensitive: there should be only
   * lower case symbols.
   * I know \B is legal in the regex.  But I prefer still
   * making the test: those who know how to use \B will have
   * no problems to understand that the test is wrong, not them */
  res &= check_words_lower_case (sheet->key, sheet->keywords);
  res &= check_words_lower_case (sheet->key, sheet->operators);
  res &= check_sequences_lower_case (sheet->key, sheet->sequences);

  return res;
}


static int
check_symbols_doubles (const char * name, struct darray * symbols)
{
  int i, res = TRUE;

  for (i = 1 ; i < symbols->len ; i++)
    if (!ustrcmp (((struct symbol *) symbols->content [i - 1])->word,
		  ((struct symbol *) symbols->content [i])->word))
      {
	res = FALSE;
	error (0, 0, "%s.ssh:`%s' is defined twice",
	       name, ((struct symbol *) symbols->content [i])->word);
      }
  return res;
}


static int
check_words_doubles (const char * name, struct words * words)
{
  return (check_symbols_doubles (name, words->strings)
	  && check_symbols_doubles (name, words->regexps));
}

/*
 * Check that there are no doublons in the style sheets
 */
static int
check_doubles (struct style_sheet * sheet)
{
  int res = TRUE;

  res &= check_words_doubles (sheet->key, sheet->keywords);
  res &= check_words_doubles (sheet->key, sheet->operators);
  
  return res;
}

/*
 * Return a correctness result upon SHEET
 * (0: failure, 1: correct)
 */
static int
style_sheet_check (struct style_sheet * sheet)
{
  return check_doubles (sheet) && check_sensitivity (sheet);
}

/************************************************************************/
/*			Handling of the hash table of			*/
/*				the style sheets			*/
/************************************************************************/
/*
 * We want to store the yet read style sheets in an hash table
 */
static unsigned long
sheet_hash_1 (void const *key)
{
  return_STRING_HASH_1 (((struct style_sheet *)key)->key);
}

static unsigned long
sheet_hash_2 (void const *key)
{
  return_STRING_HASH_2 (((struct style_sheet *)key)->key);
}

static int
sheet_hash_cmp (void const *x, void const *y)
{
  return_STRING_COMPARE (((struct style_sheet *)x)->key,
			 ((struct style_sheet *)y)->key);
}

/*
 * Create the hash table
 */
struct hash_table *
new_style_sheets (void)
{
  struct hash_table * res; 
  res = ALLOC (hash_table, 1);
  hash_init (res, 8,
             sheet_hash_1, sheet_hash_2, sheet_hash_cmp);
  return res;
}

/*
 * Make sure the style sheet SHEET is in the hash table
 */
void
require_style_sheet (const char * key)
{
  struct style_sheet * item;
  struct style_sheet token;

  if (strequ (key, "plain"))
    return;

  token.key = key;
  item = (struct style_sheet *) hash_find_item (style_sheets, &token);
  
  if (item == NULL) {
    /* The style has never been seen before, then load it
     * through a correct call to yyparse, which in turn
     * in charge to put the newly read style in the hash
     * table.  Hence, look for it again.  An error at this
     * point would be highly strange
     */
    
    /* However, as a special case, what is suppose to be a key
     * can actually be a full path to the style sheet */
    if (*key == DIR_SEP)
      item = load_style_sheet_by_file (key);
    else
      item = load_style_sheet_by_key (key);
    if (item == NULL)
      error (0, 0, _("cannot find style sheet `%s': using plain style"), key);
  }

  return;
}

/*
 * return the style sheet bearing this name
 * autoload it if not yet read.
 * If the sheet cannot be found, return NULL
 */
struct style_sheet *
get_style_sheet (const char * key)
{
  struct style_sheet * item;
  struct style_sheet token;

  if (strequ (key, "plain"))
    return NULL;

  /* Make sure the hash table contains the corresponding style sheet */
  require_style_sheet (key);

  token.key = key;
  item = (struct style_sheet *) hash_find_item (style_sheets, &token);

  if (!item)
    return NULL;

  /* To me, it looks like an horror: the syntax table is not
   * stored in the regex buffer.  As a consequence, each time
   * we change sheet, make sure to update the re_syntax_table */
  re_syntax_table = (char *) item->alpha1;
  
  return item;
}

/*
 * Called by ssh_parse to store what it read
 */
void
add_style_sheet (struct style_sheet * sheet)
{
  hash_insert (style_sheets, sheet);
}

/*
 * Compute the total SUM of the all the versions of the
 * sheets which key is in KEYS
 */
static inline void
style_sheets_versions (struct darray * keys, 
		       int sum [4], int requirement [4])
{
  int i;
  struct style_sheet * sheet;
  
  version_set_to_null (sum);
  version_set_to_null (requirement);
  for (i = 0 ; i < keys->len ; i++)
    {
      sheet = get_style_sheet ((char *) keys->content[i]);
      version_add (sum, sheet->version);
      if (version_cmp (sheet->requirement, requirement) > 0)
	version_cpy (requirement, sheet->requirement);
    }
  
}
